#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""OpenStackClient plugin for advanced Networking service."""

import logging

# TODO(rtheis/amotoki): Add functional test infrastructure for OSC
# plugin commands.
# TODO(amotoki): Add and update document on OSC plugin.

from osc_lib import utils

LOG = logging.getLogger(__name__)

DEFAULT_API_VERSION = '2.0'
API_VERSION_OPTION = 'os_network_api_version'
# NOTE(rtheis): API_NAME must NOT be set to 'network' since
# 'network' is owned by OSC!  The OSC 'network' client uses
# the OpenStack SDK.
API_NAME = 'neutronclient'
API_VERSIONS = {
    '2.0': 'neutronclient.v2_0.client.Client',
    '2': 'neutronclient.v2_0.client.Client',
}


def make_client(instance):
    """Returns an neutron client."""
    neutron_client = utils.get_client_class(
        API_NAME,
        instance._api_version[API_NAME],
        API_VERSIONS)
    LOG.debug('Instantiating neutron client: %s', neutron_client)

    # TODO(amotoki): Check the following arguments need to be passed
    # to neutronclient class. Check keystoneauth code.
    # - endpoint_type (do we need to specify it explicitly?)
    # - auth (session object contains auth. Is it required?)
    client = neutron_client(session=instance.session,
                            region_name=instance.region_name,
                            endpoint_type=instance.interface,
                            insecure=not instance.verify,
                            ca_cert=instance.cacert)
    return client


def build_option_parser(parser):
    """Hook to add global options"""

    # NOTE(amotoki): At now we register no option.
    # OSC itself has an option for Network API version # and we refer to it.
    return parser



# Copyright 2017 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock

from osc_lib.tests import utils as tests_utils

from neutronclient.osc import utils as osc_utils
from neutronclient.osc.v2.vpnaas import ikepolicy
from neutronclient.tests.unit.osc.v2 import fakes as test_fakes
from neutronclient.tests.unit.osc.v2.vpnaas import common
from neutronclient.tests.unit.osc.v2.vpnaas import fakes


_ikepolicy = fakes.IKEPolicy().create()
CONVERT_MAP = {
    'project': 'project_id',
}


def _generate_data(ordered_dict=None, data=None):
    source = ordered_dict if ordered_dict else _ikepolicy
    if data:
        source.update(data)
    return source


def _generate_req_and_res(verifylist):
    request = dict(verifylist)
    response = _ikepolicy
    for key, val in verifylist:
        converted = CONVERT_MAP.get(key, key)
        del request[key]
        new_value = val
        request[converted] = new_value
        response[converted] = new_value
    return request, response


class TestIKEPolicy(test_fakes.TestNeutronClientOSCV2):

    def check_results(self, headers, data, exp_req, is_list=False):
        if is_list:
            req_body = {self.res_plural: list(exp_req)}
        else:
            req_body = exp_req
        self.mocked.assert_called_once_with(**req_body)
        self.assertEqual(self.ordered_headers, tuple(sorted(headers)))
        self.assertEqual(self.ordered_data, data)

    def setUp(self):
        super(TestIKEPolicy, self).setUp()

        def _mock_ikepolicy(*args, **kwargs):
            self.networkclient.find_vpn_ike_policy.assert_called_once_with(
                self.resource['id'], ignore_missing=False)
            return {'id': args[0]}

        self.networkclient.find_vpn_ike_policy.side_effect = mock.Mock(
            side_effect=_mock_ikepolicy)
        osc_utils.find_project = mock.Mock()
        osc_utils.find_project.id = _ikepolicy['project_id']
        self.res = 'ikepolicy'
        self.res_plural = 'ikepolicies'
        self.resource = _ikepolicy
        self.headers = (
            'ID',
            'Name',
            'Authentication Algorithm',
            'Encryption Algorithm',
            'IKE Version',
            'Perfect Forward Secrecy (PFS)',
            'Description',
            'Phase1 Negotiation Mode',
            'Project',
            'Lifetime',
        )
        self.data = _generate_data()
        self.ordered_headers = (
            'Authentication Algorithm',
            'Description',
            'Encryption Algorithm',
            'ID',
            'IKE Version',
            'Lifetime',
            'Name',
            'Perfect Forward Secrecy (PFS)',
            'Phase1 Negotiation Mode',
            'Project',
        )
        self.ordered_data = (
            _ikepolicy['auth_algorithm'],
            _ikepolicy['description'],
            _ikepolicy['encryption_algorithm'],
            _ikepolicy['id'],
            _ikepolicy['ike_version'],
            _ikepolicy['lifetime'],
            _ikepolicy['name'],
            _ikepolicy['pfs'],
            _ikepolicy['phase1_negotiation_mode'],
            _ikepolicy['project_id'],
        )
        self.ordered_columns = (
            'auth_algorithm',
            'description',
            'encryption_algorithm',
            'id',
            'ike_version',
            'lifetime',
            'name',
            'pfs',
            'phase1_negotiation_mode',
            'project_id',
        )


class TestCreateIKEPolicy(TestIKEPolicy, common.TestCreateVPNaaS):

    def setUp(self):
        super(TestCreateIKEPolicy, self).setUp()
        self.networkclient.create_vpn_ike_policy = mock.Mock(
            return_value=_ikepolicy)
        self.mocked = self.networkclient.create_vpn_ike_policy
        self.cmd = ikepolicy.CreateIKEPolicy(self.app, self.namespace)

    def _update_expect_response(self, request, response):
        """Set expected request and response

        :param request
            A dictionary of request body(dict of verifylist)
        :param response
            A OrderedDict of request body
        """
        # Update response body
        self.networkclient.create_vpn_ikepolicy.return_value = response
        osc_utils.find_project.return_value.id = response['project_id']
        # Update response(finally returns 'data')
        self.data = _generate_data(ordered_dict=response)
        self.ordered_data = tuple(
            response[column] for column in self.ordered_columns
        )

    def _set_all_params(self, args={}):
        name = args.get('name') or 'my-name'
        description = args.get('description') or 'my-desc'
        auth_algorithm = args.get('auth_algorithm') or 'sha1'
        encryption_algorithm = args.get('encryption_algorithm') or 'aes-128'
        phase1_negotiation_mode = args.get('phase1_negotiation_mode') or 'main'
        ike_version = args.get('ike_version') or 'v1'
        pfs = args.get('pfs') or 'group5'
        tenant_id = args.get('tenant_id') or 'my-tenant'
        arglist = [
            '--description', description,
            '--auth-algorithm', auth_algorithm,
            '--encryption-algorithm', encryption_algorithm,
            '--phase1-negotiation-mode', phase1_negotiation_mode,
            '--ike-version', ike_version,
            '--pfs', pfs,
            '--project', tenant_id,
            name,
        ]
        verifylist = [
            ('description', description),
            ('auth_algorithm', auth_algorithm),
            ('encryption_algorithm', encryption_algorithm),
            ('phase1_negotiation_mode', phase1_negotiation_mode),
            ('ike_version', ike_version),
            ('pfs', pfs),
            ('project', tenant_id),
            ('name', name),
        ]
        return arglist, verifylist

    def _test_create_with_all_params(self, args={}):
        arglist, verifylist = self._set_all_params(args)
        request, response = _generate_req_and_res(verifylist)
        self._update_expect_response(request, response)
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.check_results(headers, data, request)

    def test_create_with_no_options(self):
        arglist = []
        verifylist = []

        self.assertRaises(tests_utils.ParserException, self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_create_with_all_params(self):
        self._test_create_with_all_params()

    def test_create_with_all_params_name(self):
        self._test_create_with_all_params({'name': 'new_ikepolicy'})

    def test_create_with_all_params_aggressive_mode(self):
        self._test_create_with_all_params(
            {'phase1_negotiation_mode': 'aggressive'})


class TestDeleteIKEPolicy(TestIKEPolicy, common.TestDeleteVPNaaS):

    def setUp(self):
        super(TestDeleteIKEPolicy, self).setUp()
        self.networkclient.delete_vpn_ike_policy = mock.Mock()
        self.mocked = self.networkclient.delete_vpn_ike_policy
        self.cmd = ikepolicy.DeleteIKEPolicy(self.app, self.namespace)


class TestListIKEPolicy(TestIKEPolicy):

    def setUp(self):
        super(TestListIKEPolicy, self).setUp()
        self.cmd = ikepolicy.ListIKEPolicy(self.app, self.namespace)

        self.short_header = (
            'ID',
            'Name',
            'Authentication Algorithm',
            'Encryption Algorithm',
            'IKE Version',
            'Perfect Forward Secrecy (PFS)',
        )

        self.short_data = (
            _ikepolicy['id'],
            _ikepolicy['name'],
            _ikepolicy['auth_algorithm'],
            _ikepolicy['encryption_algorithm'],
            _ikepolicy['ike_version'],
            _ikepolicy['pfs'],
        )

        self.networkclient.vpn_ike_policies = mock.Mock(
            return_value=[_ikepolicy])
        self.mocked = self.networkclient.vpn_ike_policies

    def test_list_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with()
        self.assertEqual(list(self.headers), headers)

    def test_list_with_no_option(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with()
        self.assertEqual(list(self.short_header), headers)
        self.assertEqual([self.short_data], list(data))


class TestSetIKEPolicy(TestIKEPolicy, common.TestSetVPNaaS):

    def setUp(self):
        super(TestSetIKEPolicy, self).setUp()
        self.networkclient.update_vpn_ike_policy = mock.Mock(
            return_value=_ikepolicy)
        self.mocked = self.networkclient.update_vpn_ike_policy
        self.cmd = ikepolicy.SetIKEPolicy(self.app, self.namespace)

    def test_set_auth_algorithm_with_sha256(self):
        target = self.resource['id']
        auth_algorithm = 'sha256'
        arglist = [target, '--auth-algorithm', auth_algorithm]
        verifylist = [
            (self.res, target),
            ('auth_algorithm', auth_algorithm),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            target, **{'auth_algorithm': 'sha256'})
        self.assertIsNone(result)

    def test_set_phase1_negotiation_mode_with_aggressive(self):
        target = self.resource['id']
        phase1_negotiation_mode = 'aggressive'
        arglist = [target,
                   '--phase1-negotiation-mode', phase1_negotiation_mode]
        verifylist = [
            (self.res, target),
            ('phase1_negotiation_mode', phase1_negotiation_mode),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            target, **{'phase1_negotiation_mode': 'aggressive'})
        self.assertIsNone(result)


class TestShowIKEPolicy(TestIKEPolicy, common.TestShowVPNaaS):

    def setUp(self):
        super(TestShowIKEPolicy, self).setUp()
        self.networkclient.get_vpn_ike_policy = mock.Mock(
            return_value=_ikepolicy)
        self.mocked = self.networkclient.get_vpn_ike_policy
        self.cmd = ikepolicy.ShowIKEPolicy(self.app, self.namespace)



# Copyright 2012 OpenStack Foundation.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import argparse

from oslo_serialization import jsonutils

from neutronclient._i18n import _
from neutronclient.common import exceptions
from neutronclient.common import utils
from neutronclient.neutron import v2_0 as neutronV20
from neutronclient.neutron.v2_0 import dns
from neutronclient.neutron.v2_0.qos import policy as qos_policy


def _format_fixed_ips(port):
    try:
        return '\n'.join([jsonutils.dumps(ip) for ip in port['fixed_ips']])
    except (TypeError, KeyError):
        return ''


def _add_updatable_args(parser):
    parser.add_argument(
        '--name',
        help=_('Name of this port.'))
    parser.add_argument(
        '--description',
        help=_('Description of this port.'))
    parser.add_argument(
        '--fixed-ip', metavar='subnet_id=SUBNET,ip_address=IP_ADDR',
        action='append',
        type=utils.str2dict_type(optional_keys=['subnet_id', 'ip_address']),
        help=_('Desired IP and/or subnet for this port: '
               'subnet_id=<name_or_id>,ip_address=<ip>. '
               'You can repeat this option.'))
    parser.add_argument(
        '--fixed_ip',
        action='append',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--device-id',
        help=_('Device ID of this port.'))
    parser.add_argument(
        '--device_id',
        help=argparse.SUPPRESS)
    parser.add_argument(
        '--device-owner',
        help=_('Device owner of this port.'))
    parser.add_argument(
        '--device_owner',
        help=argparse.SUPPRESS)


def _updatable_args2body(parsed_args, body, client):
    neutronV20.update_dict(parsed_args, body,
                           ['device_id', 'device_owner', 'name',
                            'description'])
    ips = []
    if parsed_args.fixed_ip:
        for ip_spec in parsed_args.fixed_ip:
            if 'subnet_id' in ip_spec:
                subnet_name_id = ip_spec['subnet_id']
                _subnet_id = neutronV20.find_resourceid_by_name_or_id(
                    client, 'subnet', subnet_name_id)
                ip_spec['subnet_id'] = _subnet_id
            ips.append(ip_spec)
    if ips:
        body['fixed_ips'] = ips


class ListPort(neutronV20.ListCommand):
    """List ports that belong to a given tenant."""

    resource = 'port'
    _formatters = {'fixed_ips': _format_fixed_ips, }
    list_columns = ['id', 'name', 'mac_address', 'fixed_ips']
    pagination_support = True
    sorting_support = True


class ListRouterPort(neutronV20.ListCommand):
    """List ports that belong to a given tenant, with specified router."""

    resource = 'port'
    _formatters = {'fixed_ips': _format_fixed_ips, }
    list_columns = ['id', 'name', 'mac_address', 'fixed_ips']
    pagination_support = True
    sorting_support = True

    def get_parser(self, prog_name):
        parser = super(ListRouterPort, self).get_parser(prog_name)
        parser.add_argument(
            'id', metavar='ROUTER',
            help=_('ID or name of the router to look up.'))
        return parser

    def take_action(self, parsed_args):
        neutron_client = self.get_client()
        _id = neutronV20.find_resourceid_by_name_or_id(
            neutron_client, 'router', parsed_args.id)
        self.values_specs.append('--device_id=%s' % _id)
        return super(ListRouterPort, self).take_action(parsed_args)


class ShowPort(neutronV20.ShowCommand):
    """Show information of a given port."""

    resource = 'port'


class UpdatePortSecGroupMixin(object):
    def add_arguments_secgroup(self, parser):
        group_sg = parser.add_mutually_exclusive_group()
        group_sg.add_argument(
            '--security-group', metavar='SECURITY_GROUP',
            default=[], action='append', dest='security_groups',
            help=_('Security group associated with the port. You can '
                   'repeat this option.'))
        group_sg.add_argument(
            '--no-security-groups',
            action='store_true',
            help=_('Associate no security groups with the port.'))

    def _resolv_sgid(self, secgroup):
        return neutronV20.find_resourceid_by_name_or_id(
            self.get_client(), 'security_group', secgroup)

    def args2body_secgroup(self, parsed_args, port):
        if parsed_args.security_groups:
            port['security_groups'] = [self._resolv_sgid(sg) for sg
                                       in parsed_args.security_groups]
        elif parsed_args.no_security_groups:
            port['security_groups'] = []


class UpdateExtraDhcpOptMixin(object):
    def add_arguments_extradhcpopt(self, parser):
        group_sg = parser.add_mutually_exclusive_group()
        group_sg.add_argument(
            '--extra-dhcp-opt',
            default=[],
            action='append',
            dest='extra_dhcp_opts',
            type=utils.str2dict_type(
                required_keys=['opt_name'],
                optional_keys=['opt_value', 'ip_version']),
            help=_('Extra dhcp options to be assigned to this port: '
                   'opt_name=<dhcp_option_name>,opt_value=<value>,'
                   'ip_version={4,6}. You can repeat this option.'))

    def args2body_extradhcpopt(self, parsed_args, port):
        ops = []
        if parsed_args.extra_dhcp_opts:
            # the extra_dhcp_opt params (opt_name & opt_value)
            # must come in pairs, if there is a parm error
            # both must be thrown out.
            opt_ele = {}
            edo_err_msg = _("Invalid --extra-dhcp-opt option, can only be: "
                            "opt_name=<dhcp_option_name>,opt_value=<value>,"
                            "ip_version={4,6}. "
                            "You can repeat this option.")
            for opt in parsed_args.extra_dhcp_opts:
                opt_ele.update(opt)
                if ('opt_name' in opt_ele and
                        ('opt_value' in opt_ele or 'ip_version' in opt_ele)):
                    if opt_ele.get('opt_value') == 'null':
                        opt_ele['opt_value'] = None
                    ops.append(opt_ele)
                    opt_ele = {}
                else:
                    raise exceptions.CommandError(edo_err_msg)

        if ops:
            port['extra_dhcp_opts'] = ops


class UpdatePortAllowedAddressPair(object):
    """Update Port for allowed_address_pairs"""

    def add_arguments_allowedaddresspairs(self, parser):
        group_aap = parser.add_mutually_exclusive_group()
        group_aap.add_argument(
            '--allowed-address-pair',
            metavar='ip_address=IP_ADDR|CIDR[,mac_address=MAC_ADDR]',
            default=[],
            action='append',
            dest='allowed_address_pairs',
            type=utils.str2dict_type(
                required_keys=['ip_address'],
                optional_keys=['mac_address']),
            help=_('Allowed address pair associated with the port. '
                   '"ip_address" parameter is required. IP address or '
                   'CIDR can be specified for "ip_address". '
                   '"mac_address" parameter is optional. '
                   'You can repeat this option.'))
        group_aap.add_argument(
            '--no-allowed-address-pairs',
            action='store_true',
            help=_('Associate no allowed address pairs with the port.'))

    def args2body_allowedaddresspairs(self, parsed_args, port):
        if parsed_args.allowed_address_pairs:
            port['allowed_address_pairs'] = parsed_args.allowed_address_pairs
        elif parsed_args.no_allowed_address_pairs:
            port['allowed_address_pairs'] = []


class CreatePort(neutronV20.CreateCommand, UpdatePortSecGroupMixin,
                 UpdateExtraDhcpOptMixin, qos_policy.CreateQosPolicyMixin,
                 UpdatePortAllowedAddressPair):
    """Create a port for a given tenant."""

    resource = 'port'

    def add_known_arguments(self, parser):
        _add_updatable_args(parser)
        parser.add_argument(
            '--admin-state-down',
            dest='admin_state', action='store_false',
            help=_('Set admin state up to false.'))
        parser.add_argument(
            '--admin_state_down',
            dest='admin_state', action='store_false',
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--mac-address',
            help=_('MAC address of this port.'))
        parser.add_argument(
            '--mac_address',
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--vnic-type',
            metavar='<direct | direct-physical | macvtap '
                    '| normal | baremetal | smart-nic>',
            choices=['direct', 'direct-physical', 'macvtap',
                     'normal', 'baremetal', 'smart-nic'],
            type=utils.convert_to_lowercase,
            help=_('VNIC type for this port.'))
        parser.add_argument(
            '--vnic_type',
            choices=['direct', 'direct-physical', 'macvtap',
                     'normal', 'baremetal', 'smart-nic'],
            type=utils.convert_to_lowercase,
            help=argparse.SUPPRESS)
        parser.add_argument(
            '--binding-profile',
            help=_('Custom data to be passed as binding:profile.'))
        parser.add_argument(
            '--binding_profile',
            help=argparse.SUPPRESS)
        self.add_arguments_secgroup(parser)
        self.add_arguments_extradhcpopt(parser)
        self.add_arguments_qos_policy(parser)
        self.add_arguments_allowedaddresspairs(parser)

        parser.add_argument(
            'network_id', metavar='NETWORK',
            help=_('ID or name of the network this port belongs to.'))
        dns.add_dns_argument_create(parser, self.resource, 'name')

    def args2body(self, parsed_args):
        client = self.get_client()
        _network_id = neutronV20.find_resourceid_by_name_or_id(
            client, 'network', parsed_args.network_id)
        body = {'admin_state_up': parsed_args.admin_state,
                'network_id': _network_id, }
        _updatable_args2body(parsed_args, body, client)
        neutronV20.update_dict(parsed_args, body,
                               ['mac_address', 'tenant_id'])
        if parsed_args.vnic_type:
            body['binding:vnic_type'] = parsed_args.vnic_type
        if parsed_args.binding_profile:
            body['binding:profile'] = jsonutils.loads(
                parsed_args.binding_profile)

        self.args2body_secgroup(parsed_args, body)
        self.args2body_extradhcpopt(parsed_args, body)
        self.args2body_qos_policy(parsed_args, body)
        self.args2body_allowedaddresspairs(parsed_args, body)
        dns.args2body_dns_create(parsed_args, body, 'name')

        return {'port': body}


class DeletePort(neutronV20.DeleteCommand):
    """Delete a given port."""

    resource = 'port'


class UpdatePort(neutronV20.UpdateCommand, UpdatePortSecGroupMixin,
                 UpdateExtraDhcpOptMixin, qos_policy.UpdateQosPolicyMixin,
                 UpdatePortAllowedAddressPair):
    """Update port's information."""

    resource = 'port'

    def add_known_arguments(self, parser):
        _add_updatable_args(parser)
        parser.add_argument(
            '--admin-state-up',
            choices=['True', 'False'],
            help=_('Set admin state up for the port.'))
        parser.add_argument(
            '--admin_state_up',
            choices=['True', 'False'],
            help=argparse.SUPPRESS)
        self.add_arguments_secgroup(parser)
        self.add_arguments_extradhcpopt(parser)
        self.add_arguments_qos_policy(parser)
        self.add_arguments_allowedaddresspairs(parser)
        dns.add_dns_argument_update(parser, self.resource, 'name')

    def args2body(self, parsed_args):
        body = {}
        client = self.get_client()
        _updatable_args2body(parsed_args, body, client)
        if parsed_args.admin_state_up:
            body['admin_state_up'] = parsed_args.admin_state_up

        self.args2body_secgroup(parsed_args, body)
        self.args2body_extradhcpopt(parsed_args, body)
        self.args2body_qos_policy(parsed_args, body)
        self.args2body_allowedaddresspairs(parsed_args, body)
        dns.args2body_dns_update(parsed_args, body, 'name')

        return {'port': body}



# Copyright 2017 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import collections
import copy
from unittest import mock
import uuid


class FakeLogging(object):

    def create(self, attrs={}):
        """Create a fake network logs

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A OrderedDict faking the network log
        """
        self.ordered.update(attrs)
        return copy.deepcopy(self.ordered)

    def bulk_create(self, attrs=None, count=2):
        """Create multiple fake network logs

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of network logs to fake
        :return:
            A list of dictionaries faking the network logs
        """
        return [self.create(attrs=attrs) for i in range(0, count)]

    def get(self, attrs=None, count=2):
        """Create multiple fake network logs

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of network logs to fake
        :return:
            A list of dictionaries faking the network log
        """
        if attrs is None:
            self.attrs = self.bulk_create(count=count)
        return mock.Mock(side_effect=attrs)


class NetworkLog(FakeLogging):
    """Fake one or more network log"""

    def __init__(self):
        super(NetworkLog, self).__init__()
        self.ordered = collections.OrderedDict((
            ('id', 'log-id-' + uuid.uuid4().hex),
            ('description', 'my-desc-' + uuid.uuid4().hex),
            ('enabled', False),
            ('name', 'my-log-' + uuid.uuid4().hex),
            ('target_id', None),
            ('project_id', 'project-id-' + uuid.uuid4().hex),
            ('resource_id', None),
            ('resource_type', 'security_group'),
            ('event', 'all'),
        ))



# Copyright 2014 Blue Box Group, Inc.
# Copyright 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutronclient._i18n import _
from neutronclient.common import exceptions
from neutronclient.common import utils
from neutronclient.neutron import v2_0 as neutronV20


def _get_loadbalancer_id(client, lb_id_or_name):
    return neutronV20.find_resourceid_by_name_or_id(
        client, 'loadbalancer', lb_id_or_name,
        cmd_resource='lbaas_loadbalancer')


def _get_pool(client, pool_id_or_name):
    return neutronV20.find_resource_by_name_or_id(
        client, 'pool', pool_id_or_name, cmd_resource='lbaas_pool')


def _get_pool_id(client, pool_id_or_name):
    return neutronV20.find_resourceid_by_name_or_id(
        client, 'pool', pool_id_or_name, cmd_resource='lbaas_pool')


def _add_common_args(parser):
    parser.add_argument(
        '--description',
        help=_('Description of the listener.'))
    parser.add_argument(
        '--connection-limit',
        type=int,
        help=_('The maximum number of connections per second allowed for '
               'the listener. Positive integer or -1 '
               'for unlimited (default).'))
    parser.add_argument(
        '--default-pool',
        help=_('Default pool for the listener.'))


def _parse_common_args(body, parsed_args, client):
    neutronV20.update_dict(parsed_args, body,
                           ['name', 'description', 'connection_limit'])
    if parsed_args.default_pool:
        default_pool_id = _get_pool_id(
            client, parsed_args.default_pool)
        body['default_pool_id'] = default_pool_id


class ListListener(neutronV20.ListCommand):
    """LBaaS v2 List listeners that belong to a given tenant."""

    resource = 'listener'
    list_columns = ['id', 'default_pool_id', 'name', 'protocol',
                    'protocol_port', 'admin_state_up', 'status']
    pagination_support = True
    sorting_support = True


class ShowListener(neutronV20.ShowCommand):
    """LBaaS v2 Show information of a given listener."""

    resource = 'listener'


class CreateListener(neutronV20.CreateCommand):
    """LBaaS v2 Create a listener."""

    resource = 'listener'

    def add_known_arguments(self, parser):
        _add_common_args(parser)
        parser.add_argument(
            '--admin-state-down',
            dest='admin_state', action='store_false',
            help=_('Set admin state up to false.'))
        parser.add_argument(
            '--name',
            help=_('The name of the listener. At least one of --default-pool '
                   'or --loadbalancer must be specified.'))
        parser.add_argument(
            '--default-tls-container-ref',
            dest='default_tls_container_ref',
            help=_('Default TLS container reference'
                   ' to retrieve TLS information.'))
        parser.add_argument(
            '--sni-container-refs',
            dest='sni_container_refs',
            nargs='+',
            help=_('List of TLS container references for SNI.'))
        parser.add_argument(
            '--loadbalancer',
            metavar='LOADBALANCER',
            help=_('ID or name of the load balancer.'))
        parser.add_argument(
            '--protocol',
            required=True,
            choices=['TCP', 'HTTP', 'HTTPS', 'TERMINATED_HTTPS'],
            type=utils.convert_to_uppercase,
            help=_('Protocol for the listener.'))
        parser.add_argument(
            '--protocol-port',
            dest='protocol_port', required=True,
            metavar='PORT',
            help=_('Protocol port for the listener.'))

    def args2body(self, parsed_args):
        if not parsed_args.loadbalancer and not parsed_args.default_pool:
            message = _('Either --default-pool or --loadbalancer must be '
                        'specified.')
            raise exceptions.CommandError(message)
        body = {
            'protocol': parsed_args.protocol,
            'protocol_port': parsed_args.protocol_port,
            'admin_state_up': parsed_args.admin_state
        }
        if parsed_args.loadbalancer:
            loadbalancer_id = _get_loadbalancer_id(
                self.get_client(), parsed_args.loadbalancer)
            body['loadbalancer_id'] = loadbalancer_id

        neutronV20.update_dict(parsed_args, body,
                               ['default_tls_container_ref',
                                'sni_container_refs', 'tenant_id'])
        _parse_common_args(body, parsed_args, self.get_client())
        return {self.resource: body}


class UpdateListener(neutronV20.UpdateCommand):
    """LBaaS v2 Update a given listener."""

    resource = 'listener'

    def add_known_arguments(self, parser):
        _add_common_args(parser)
        parser.add_argument(
            '--name',
            help=_('Name of the listener.'))
        utils.add_boolean_argument(
            parser, '--admin-state-up', dest='admin_state_up',
            help=_('Specify the administrative state of the listener. '
                   '(True meaning "Up")'))

    def args2body(self, parsed_args):
        body = {}
        neutronV20.update_dict(parsed_args, body,
                               ['admin_state_up'])
        _parse_common_args(body, parsed_args, self.get_client())
        return {self.resource: body}


class DeleteListener(neutronV20.DeleteCommand):
    """LBaaS v2 Delete a given listener."""

    resource = 'listener'



# Copyright (c) 2017 Huawei Technologies India Pvt.Limited.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
from unittest import mock


from osc_lib.tests import utils
from oslo_utils import uuidutils

from openstack.network.v2 import sfc_flow_classifier as flow_classifier
from openstack.network.v2 import sfc_port_chain as port_chain
from openstack.network.v2 import sfc_port_pair as port_pair
from openstack.network.v2 import sfc_port_pair_group as port_pair_group
from openstack.network.v2 import sfc_service_graph as service_graph


class TestNeutronClientOSCV2(utils.TestCommand):

    def setUp(self):
        super(TestNeutronClientOSCV2, self).setUp()
        self.namespace = argparse.Namespace()
        self.app.client_manager.session = mock.Mock()
        self.app.client_manager.network = mock.Mock()
        self.network = self.app.client_manager.network
        self.network.find_sfc_flow_classifier = mock.Mock(
            side_effect=lambda name_or_id, ignore_missing=False:
            {'id': name_or_id}
        )
        self.network.find_sfc_port_chain = mock.Mock(
            side_effect=lambda name_or_id, ignore_missing=False:
            {'id': name_or_id}
        )
        self.network.find_sfc_port_pair = mock.Mock(
            side_effect=lambda name_or_id, ignore_missing=False:
            {'id': name_or_id}
        )
        self.network.find_sfc_port_pair_group = mock.Mock(
            side_effect=lambda name_or_id, ignore_missing=False:
            {'id': name_or_id}
        )
        self.network.find_sfc_service_graph = mock.Mock(
            side_effect=lambda name_or_id, ignore_missing=False:
            {'id': name_or_id}
        )
        self.network.find_port = mock.Mock(
            side_effect=lambda name_or_id, ignore_missing=False:
            {'id': name_or_id}
        )


class FakeSfcPortPair(object):
    """Fake port pair attributes."""

    @staticmethod
    def create_port_pair(attrs=None):
        """Create a fake port pair.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A Dictionary with id, name, description, ingress, egress,
            service-function-parameter, project_id
        """
        attrs = attrs or {}

        # Set default attributes.
        port_pair_attrs = {
            'description': 'description',
            'egress': uuidutils.generate_uuid(),
            'id': uuidutils.generate_uuid(),
            'ingress': uuidutils.generate_uuid(),
            'name': 'port-pair-name',
            'service_function_parameters': [('correlation', None),
                                            ('weight', 1)],
            'project_id': uuidutils.generate_uuid(),
        }

        # Overwrite default attributes.
        port_pair_attrs.update(attrs)
        return port_pair.SfcPortPair(**port_pair_attrs)

    @staticmethod
    def create_port_pairs(attrs=None, count=1):
        """Create multiple port_pairs.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of port_pairs to fake
        :return:
            A list of dictionaries faking the port_pairs
        """
        port_pairs = []
        for _ in range(count):
            port_pairs.append(FakeSfcPortPair.create_port_pair(attrs))

        return port_pairs


class FakeSfcPortPairGroup(object):
    """Fake port pair group attributes."""

    @staticmethod
    def create_port_pair_group(attrs=None):
        """Create a fake port pair group.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A Dictionary with id, name, description, port_pairs, group_id
            port_pair_group_parameters, project_id
        """
        attrs = attrs or {}

        # Set default attributes.
        port_pair_group_attrs = {
            'id': uuidutils.generate_uuid(),
            'name': 'port-pair-group-name',
            'description': 'description',
            'port_pairs': uuidutils.generate_uuid(),
            'port_pair_group_parameters': {"lb_fields": []},
            'project_id': uuidutils.generate_uuid(),
            'tap_enabled': False
        }

        port_pair_group_attrs.update(attrs)
        return port_pair_group.SfcPortPairGroup(**port_pair_group_attrs)

    @staticmethod
    def create_port_pair_groups(attrs=None, count=1):
        """Create multiple port pair groups.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of port_pair_groups to fake
        :return:
            A list of dictionaries faking the port pair groups
        """
        port_pair_groups = []
        for _ in range(count):
            port_pair_groups.append(
                FakeSfcPortPairGroup.create_port_pair_group(attrs))

        return port_pair_groups


class FakeSfcFlowClassifier(object):
    """Fake flow classifier attributes."""

    @staticmethod
    def create_flow_classifier(attrs=None):
        """Create a fake flow classifier.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A Dictionary with faking port chain attributes
        """
        attrs = attrs or {}

        # Set default attributes.
        flow_classifier_attrs = {
            'id': uuidutils.generate_uuid(),
            'destination_ip_prefix': '2.2.2.2/32',
            'destination_port_range_max': '90',
            'destination_port_range_min': '80',
            'ethertype': 'IPv4',
            'logical_destination_port': uuidutils.generate_uuid(),
            'logical_source_port': uuidutils.generate_uuid(),
            'name': 'flow-classifier-name',
            'description': 'fc_description',
            'protocol': 'tcp',
            'source_ip_prefix': '1.1.1.1/32',
            'source_port_range_max': '20',
            'source_port_range_min': '10',
            'project_id': uuidutils.generate_uuid(),
            'l7_parameters': {}
        }
        flow_classifier_attrs.update(attrs)
        return flow_classifier.SfcFlowClassifier(**flow_classifier_attrs)

    @staticmethod
    def create_flow_classifiers(attrs=None, count=1):
        """Create multiple flow classifiers.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of flow classifiers to fake
        :return:
            A list of dictionaries faking the flow classifiers
        """
        flow_classifiers = []
        for _ in range(count):
            flow_classifiers.append(
                FakeSfcFlowClassifier.create_flow_classifier(attrs))

        return flow_classifiers


class FakeSfcPortChain(object):
    """Fake port chain attributes."""

    @staticmethod
    def create_port_chain(attrs=None):
        """Create a fake port chain.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A Dictionary with faking port chain attributes
        """
        attrs = attrs or {}

        # Set default attributes.
        port_chain_attrs = {
            'id': uuidutils.generate_uuid(),
            'name': 'port-chain-name',
            'description': 'description',
            'port_pair_groups': uuidutils.generate_uuid(),
            'flow_classifiers': uuidutils.generate_uuid(),
            'chain_parameters': {"correlation": "mpls", "symmetric": False},
            'project_id': uuidutils.generate_uuid(),
        }

        port_chain_attrs.update(attrs)
        return port_chain.SfcPortChain(**port_chain_attrs)

    @staticmethod
    def create_port_chains(attrs=None, count=1):
        """Create multiple port chains.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of port chains to fake
        :return:
            A list of dictionaries faking the port chains.
        """
        port_chains = []
        for _ in range(count):
            port_chains.append(FakeSfcPortChain.create_port_chain(attrs))
        return port_chains


class FakeSfcServiceGraph(object):
    """Fake service graph attributes."""

    @staticmethod
    def create_sfc_service_graph(attrs=None):
        """Create a fake service graph.

        :param Dictionary attrs:
            A dictionary with all attributes
        :return:
            A Dictionary with faking service graph attributes
        """
        attrs = attrs or {}

        # Set default attributes.
        service_graph_attrs = {
            'id': uuidutils.generate_uuid(),
            'name': 'port-pair-group-name',
            'description': 'description',
            'port_chains': {uuidutils.generate_uuid(): [
                uuidutils.generate_uuid()]},
            'project_id': uuidutils.generate_uuid(),
        }

        service_graph_attrs.update(attrs)
        return service_graph.SfcServiceGraph(**service_graph_attrs)

    @staticmethod
    def create_sfc_service_graphs(attrs=None, count=1):
        """Create multiple service graphs.

        :param Dictionary attrs:
            A dictionary with all attributes
        :param int count:
            The number of service graphs to fake
        :return:
            A list of dictionaries faking the service graphs.
        """
        service_graphs = []
        for _ in range(count):
            service_graphs.append(
                FakeSfcServiceGraph.create_sfc_service_graph(attrs))
        return service_graphs



# Copyright 2017 FUJITSU LIMITED
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from unittest import mock
import uuid


from neutronclient.osc import utils as osc_utils
from neutronclient.osc.v2.vpnaas import vpnservice
from neutronclient.tests.unit.osc.v2 import fakes as test_fakes
from neutronclient.tests.unit.osc.v2.vpnaas import common
from neutronclient.tests.unit.osc.v2.vpnaas import fakes


_vpnservice = fakes.VPNService().create()
CONVERT_MAP = {
    'project': 'project_id',
    'router': 'router_id',
    'subnet': 'subnet_id'
}


def _generate_data(ordered_dict=None, data=None):
    source = ordered_dict if ordered_dict else _vpnservice
    if data:
        source.update(data)
    return source


def _generate_req_and_res(verifylist):
    request = dict(verifylist)
    response = _vpnservice
    for key, val in verifylist:
        converted = CONVERT_MAP.get(key, key)
        del request[key]
        new_value = val
        request[converted] = new_value
        response[converted] = new_value
    return request, response


class TestVPNService(test_fakes.TestNeutronClientOSCV2):

    def _check_results(self, headers, data, exp_req, is_list=False):
        if is_list:
            req_body = {self.res_plural: list(exp_req)}
        else:
            req_body = exp_req
        self.mocked.assert_called_once_with(**req_body)
        self.assertEqual(self.ordered_headers, headers)
        self.assertEqual(self.ordered_data, data)

    def setUp(self):
        super(TestVPNService, self).setUp()

        def _mock_vpnservice(*args, **kwargs):
            self.networkclient.find_vpn_service.assert_called_once_with(
                self.resource['id'], ignore_missing=False)
            return {'id': args[0]}

        self.networkclient.find_router = mock.Mock()
        self.networkclient.find_subnet = mock.Mock()
        self.fake_router = mock.Mock()
        self.fake_subnet = mock.Mock()
        self.networkclient.find_router.return_value = self.fake_router
        self.networkclient.find_subnet.return_value = self.fake_subnet
        self.args = {
            'name': 'my-name',
            'description': 'my-desc',
            'project_id': 'project-id-' + uuid.uuid4().hex,
            'router_id': 'router-id-' + uuid.uuid4().hex,
            'subnet_id': 'subnet-id-' + uuid.uuid4().hex,

        }
        self.fake_subnet.id = self.args['subnet_id']
        self.fake_router.id = self.args['router_id']

        self.networkclient.find_vpn_service.side_effect = mock.Mock(
            side_effect=_mock_vpnservice)
        osc_utils.find_project = mock.Mock()
        osc_utils.find_project.id = _vpnservice['project_id']

        self.res = 'vpnservice'
        self.res_plural = 'vpnservices'
        self.resource = _vpnservice
        self.headers = (
            'ID',
            'Name',
            'Router',
            'Subnet',
            'Flavor',
            'State',
            'Status',
            'Description',
            'Project',
            'Ext v4 IP',
            'Ext v6 IP',
        )
        self.data = _generate_data()
        self.ordered_headers = (
            'Description',
            'Ext v4 IP',
            'Ext v6 IP',
            'Flavor',
            'ID',
            'Name',
            'Project',
            'Router',
            'State',
            'Status',
            'Subnet',
        )
        self.ordered_data = (
            _vpnservice['description'],
            _vpnservice['external_v4_ip'],
            _vpnservice['external_v6_ip'],
            _vpnservice['flavor_id'],
            _vpnservice['id'],
            _vpnservice['name'],
            _vpnservice['project_id'],
            _vpnservice['router_id'],
            _vpnservice['admin_state_up'],
            _vpnservice['status'],
            _vpnservice['subnet_id'],
        )
        self.ordered_columns = (
            'description',
            'external_v4_ip',
            'external_v6_ip',
            'flavor_id',
            'id',
            'name',
            'project_id',
            'router_id',
            'admin_state_up',
            'status',
            'subnet_id',
        )


class TestCreateVPNService(TestVPNService, common.TestCreateVPNaaS):

    def setUp(self):
        super(TestCreateVPNService, self).setUp()
        self.networkclient.create_vpn_service = mock.Mock(
            return_value=_vpnservice)
        self.mocked = self.networkclient.create_vpn_service
        self.cmd = vpnservice.CreateVPNService(self.app, self.namespace)

    def _update_expect_response(self, request, response):
        """Set expected request and response

        :param request
            A dictionary of request body(dict of verifylist)
        :param response
            A OrderedDict of request body
        """
        # Update response body
        self.networkclient.create_vpn_service.return_value = response
        osc_utils.find_project.return_value.id = response['project_id']
        # Update response(finally returns 'data')
        self.data = _generate_data(ordered_dict=response)
        self.ordered_data = tuple(
            response[column] for column in self.ordered_columns
        )

    def _set_all_params(self):
        name = self.args.get('name')
        description = self.args.get('description')
        router_id = self.args.get('router_id')
        subnet_id = self.args.get('subnet_id')
        project_id = self.args.get('project_id')
        arglist = [
            '--description', description,
            '--project', project_id,
            '--subnet', subnet_id,
            '--router', router_id,
            name,
        ]
        verifylist = [
            ('description', description),
            ('project', project_id),
            ('subnet', subnet_id),
            ('router', router_id),
            ('name', name),
        ]
        return arglist, verifylist

    def _test_create_with_all_params(self):
        arglist, verifylist = self._set_all_params()
        request, response = _generate_req_and_res(verifylist)
        self._update_expect_response(request, response)
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self._check_results(headers, data, request)

    def test_create_with_all_params(self):
        self._test_create_with_all_params()


class TestDeleteVPNService(TestVPNService, common.TestDeleteVPNaaS):

    def setUp(self):
        super(TestDeleteVPNService, self).setUp()
        self.networkclient.delete_vpn_service = mock.Mock()
        self.mocked = self.networkclient.delete_vpn_service
        self.cmd = vpnservice.DeleteVPNService(self.app, self.namespace)


class TestListVPNService(TestVPNService):

    def setUp(self):
        super(TestListVPNService, self).setUp()
        self.cmd = vpnservice.ListVPNService(self.app, self.namespace)

        self.short_header = (
            'ID',
            'Name',
            'Router',
            'Subnet',
            'Flavor',
            'State',
            'Status',
        )

        self.short_data = (
            _vpnservice['id'],
            _vpnservice['name'],
            _vpnservice['router_id'],
            _vpnservice['subnet_id'],
            _vpnservice['flavor_id'],
            _vpnservice['admin_state_up'],
            _vpnservice['status'],
        )

        self.networkclient.vpn_services = mock.Mock(return_value=[_vpnservice])
        self.mocked = self.networkclient.vpn_services

    def test_list_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with()
        self.assertEqual(list(self.headers), headers)

    def test_list_with_no_option(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        headers, data = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with()
        self.assertEqual(list(self.short_header), headers)
        self.assertEqual([self.short_data], list(data))


class TestSetVPNService(TestVPNService, common.TestSetVPNaaS):

    def setUp(self):
        super(TestSetVPNService, self).setUp()
        self.networkclient.update_vpn_service = mock.Mock(
            return_value=_vpnservice)
        self.mocked = self.networkclient.update_vpn_service
        self.cmd = vpnservice.SetVPNSercice(self.app, self.namespace)

    def test_set_name(self):
        target = self.resource['id']
        update = 'change'
        arglist = [target, '--name', update]
        verifylist = [
            (self.res, target),
            ('name', update),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.mocked.assert_called_once_with(
            target, **{'name': update})
        self.assertIsNone(result)


class TestShowVPNService(TestVPNService, common.TestShowVPNaaS):

    def setUp(self):
        super(TestShowVPNService, self).setUp()
        self.networkclient.get_vpn_service = mock.Mock(
            return_value=_vpnservice)
        self.mocked = self.networkclient.get_vpn_service
        self.cmd = vpnservice.ShowVPNService(self.app, self.namespace)



# Copyright 2017 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from osc_lib import exceptions
from osc_lib.tests import utils as tests_utils
import testtools

from neutronclient.osc.v2.sfc import sfc_service_graph
from neutronclient.tests.unit.osc.v2.sfc import fakes


class TestListSfcServiceGraph(fakes.TestNeutronClientOSCV2):
    _service_graphs = fakes.FakeSfcServiceGraph.create_sfc_service_graphs(
        count=1)
    columns = ('ID', 'Name', 'Branching Points')
    columns_long = ('ID', 'Name', 'Branching Points', 'Description', 'Project')
    _service_graph = _service_graphs[0]
    data = [
        _service_graph['id'],
        _service_graph['name'],
        _service_graph['port_chains']
    ]
    data_long = [
        _service_graph['id'],
        _service_graph['name'],
        _service_graph['port_chains'],
        _service_graph['description'],
        _service_graph['project_id']
    ]
    _service_graph1 = {'service_graphs': _service_graph}
    _service_graph_id = _service_graph['id']

    def setUp(self):
        super(TestListSfcServiceGraph, self).setUp()
        self.network.sfc_service_graphs = mock.Mock(
            return_value=self._service_graphs
        )
        # Get the command object to test
        self.cmd = sfc_service_graph.ListSfcServiceGraph(
            self.app, self.namespace)

    def test_list_sfc_service_graphs(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns = self.cmd.take_action(parsed_args)[0]
        sgs = self.network.sfc_service_graphs()
        sg = sgs[0]
        data = [
            sg['id'],
            sg['name'],
            sg['port_chains']
        ]
        self.assertEqual(list(self.columns), columns)
        self.assertEqual(self.data, data)

    def test_list_sfc_service_graphs_with_long_option(self):
        arglist = ['--long']
        verifylist = [('long', True)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns = self.cmd.take_action(parsed_args)[0]
        sgs = self.network.sfc_service_graphs()
        sg = sgs[0]
        data = [
            sg['id'],
            sg['name'],
            sg['port_chains'],
            sg['description'],
            sg['project_id']
        ]
        self.assertEqual(list(self.columns_long), columns)
        self.assertEqual(self.data_long, data)


class TestCreateSfcServiceGraph(fakes.TestNeutronClientOSCV2):
    _service_graph = fakes.FakeSfcServiceGraph.create_sfc_service_graph()

    columns = ('ID', 'Name', 'Branching Points')
    columns_long = ('Branching Points', 'Description', 'ID', 'Name', 'Project')

    def get_data(self):
        return (
            self._service_graph['port_chains'],
            self._service_graph['description'],
            self._service_graph['id'],
            self._service_graph['name'],
            self._service_graph['project_id'],
        )

    def setUp(self):
        super(TestCreateSfcServiceGraph, self).setUp()
        self.network.create_sfc_service_graph = mock.Mock(
            return_value=self._service_graph)
        self.data = self.get_data()
        self.cmd = sfc_service_graph.CreateSfcServiceGraph(
            self.app, self.namespace)

    def test_create_sfc_service_graph(self):
        arglist = []
        verifylist = []

        self.assertRaises(tests_utils.ParserException, self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_create_sfc_service_graph_without_loop(self):
        bp1_str = 'pc1:pc2,pc3'
        bp2_str = 'pc2:pc4'
        self.cmd = sfc_service_graph.CreateSfcServiceGraph(
            self.app, self.namespace)

        arglist = [
            "--description", self._service_graph['description'],
            "--branching-point", bp1_str,
            "--branching-point", bp2_str,
            self._service_graph['name']]

        pcs = {'pc1': ['pc2', 'pc3'], 'pc2': ['pc4']}

        verifylist = [
            ("description", self._service_graph['description']),
            ("branching_points", [bp1_str, bp2_str]),
            ("name", self._service_graph['name'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = (self.cmd.take_action(parsed_args))

        self.network.create_sfc_service_graph.assert_called_once_with(**{
                'description': self._service_graph['description'],
                'name': self._service_graph['name'],
                'port_chains': pcs
        })
        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data, data)

    def test_create_sfc_service_graph_with_loop(self):
        bp1_str = 'pc1:pc2,pc3;'
        bp2_str = 'pc2:pc1'
        self.cmd = sfc_service_graph.CreateSfcServiceGraph(
            self.app, self.namespace)

        arglist = [
            "--description", self._service_graph['description'],
            "--branching-point", bp1_str,
            "--branching-point", bp2_str,
            self._service_graph['name']]

        verifylist = [
            ("description", self._service_graph['description']),
            ("branching_points", [bp1_str, bp2_str]),
            ("name", self._service_graph['name'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)

    def test_create_sfc_service_graph_invalid_port_chains(self):
        bp1_str = 'pc1:pc2,pc3:'
        self.cmd = sfc_service_graph.CreateSfcServiceGraph(
            self.app, self.namespace)

        arglist = [
            "--description", self._service_graph['description'],
            "--branching-point", bp1_str,
            self._service_graph['name']]

        verifylist = [
            ("description", self._service_graph['description']),
            ("branching_points", [bp1_str]),
            ("name", self._service_graph['name'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)

    def test_create_sfc_service_graph_duplicate_src_chains(self):
        bp1_str = 'pc1:pc2,pc3;'
        bp2_str = 'pc1:pc4'
        self.cmd = sfc_service_graph.CreateSfcServiceGraph(
            self.app, self.namespace)

        arglist = [
            "--description", self._service_graph['description'],
            "--branching-point", bp1_str,
            "--branching-point", bp2_str,
            self._service_graph['name']]

        verifylist = [
            ("description", self._service_graph['description']),
            ("branching_points", [bp1_str, bp2_str]),
            ("name", self._service_graph['name'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args)


class TestDeleteSfcServiceGraph(fakes.TestNeutronClientOSCV2):

    _service_graph = fakes.FakeSfcServiceGraph.create_sfc_service_graphs(
        count=1)

    def setUp(self):
        super(TestDeleteSfcServiceGraph, self).setUp()
        self.network.delete_sfc_service_graph = mock.Mock(
            return_value=None)
        self.cmd = sfc_service_graph.DeleteSfcServiceGraph(
            self.app, self.namespace)

    def test_delete_sfc_service_graph(self):
        client = self.app.client_manager.network
        mock_service_graph_delete = client.delete_sfc_service_graph
        arglist = [
            self._service_graph[0]['id'],
        ]
        verifylist = [
            ('service_graph', [self._service_graph[0]['id']]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        mock_service_graph_delete.assert_called_once_with(
            self._service_graph[0]['id'])
        self.assertIsNone(result)

    def test_delete_multiple_service_graphs_with_exception(self):
        client = self.app.client_manager.network
        target = self._service_graph[0]['id']
        arglist = [target]
        verifylist = [('service_graph', [target])]

        client.find_sfc_service_graph.side_effect = [
            target, exceptions.CommandError
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        msg = "1 of 2 service graph(s) failed to delete."
        with testtools.ExpectedException(exceptions.CommandError) as e:
            self.cmd.take_action(parsed_args)
            self.assertEqual(msg, str(e))


class TestShowSfcServiceGraph(fakes.TestNeutronClientOSCV2):

    _sg = fakes.FakeSfcServiceGraph.create_sfc_service_graph()
    columns = ('ID', 'Name', 'Branching Points')
    columns_long = ('Branching Points', 'Description', 'ID', 'Name', 'Project')
    data = (
        _sg['id'],
        _sg['name'],
        _sg['port_chains']
    )
    data_long = (
        _sg['port_chains'],
        _sg['description'],
        _sg['id'],
        _sg['name'],
        _sg['project_id']
    )

    _service_graph = _sg
    _service_graph_id = _sg['id']

    def setUp(self):
        super(TestShowSfcServiceGraph, self).setUp()
        self.network.get_sfc_service_graph = mock.Mock(
            return_value=self._service_graph
        )
        # Get the command object to test
        self.cmd = sfc_service_graph.ShowSfcServiceGraph(
            self.app, self.namespace)

    def test_service_graph_show(self):
        client = self.app.client_manager.network
        mock_service_graph_show = client.get_sfc_service_graph
        arglist = [
            self._service_graph_id,
        ]
        verifylist = [
            ('service_graph', self._service_graph_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        mock_service_graph_show.assert_called_once_with(self._service_graph_id)
        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data_long, data)


class TestSetSfcServiceGraph(fakes.TestNeutronClientOSCV2):
    _service_graph = fakes.FakeSfcServiceGraph.create_sfc_service_graph()
    _service_graph_name = _service_graph['name']
    _service_graph_id = _service_graph['id']

    def setUp(self):
        super(TestSetSfcServiceGraph, self).setUp()
        self.network.update_sfc_service_graph = mock.Mock(
            return_value=None)
        self.cmd = sfc_service_graph.SetSfcServiceGraph(
            self.app, self.namespace)

    def test_set_service_graph(self):
        client = self.app.client_manager.network
        mock_service_graph_update = client.update_sfc_service_graph
        arglist = [
            self._service_graph_name,
            '--name', 'name_updated',
            '--description', 'desc_updated'
        ]
        verifylist = [
            ('service_graph', self._service_graph_name),
            ('name', 'name_updated'),
            ('description', 'desc_updated'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'name': 'name_updated',
            'description': 'desc_updated'
        }
        mock_service_graph_update.assert_called_once_with(
            self._service_graph_name, **attrs)
        self.assertIsNone(result)



# Copyright 2013 Big Switch Networks
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import argparse

from neutronclient._i18n import _
from neutronclient.common import utils
from neutronclient.neutron import v2_0 as neutronv20


def _format_firewall_rules(firewall_policy):
    try:
        output = '[' + ',\n '.join([rule for rule in
                                    firewall_policy['firewall_rules']]) + ']'
        return output
    except (TypeError, KeyError):
        return ''


def add_common_args(parser):
    parser.add_argument(
        '--description',
        help=_('Description for the firewall policy.'))
    parser.add_argument(
        '--firewall-rules', type=lambda x: x.split(),
        help=_('Ordered list of whitespace-delimited firewall rule '
               'names or IDs; e.g., --firewall-rules \"rule1 rule2\"'))


def parse_common_args(client, parsed_args):
    if parsed_args.firewall_rules:
        _firewall_rules = []
        for f in parsed_args.firewall_rules:
            _firewall_rules.append(
                neutronv20.find_resourceid_by_name_or_id(
                    client, 'firewall_rule', f))
        body = {'firewall_rules': _firewall_rules}
    else:
        body = {}
    neutronv20.update_dict(parsed_args, body,
                           ['name', 'description', 'shared',
                            'audited', 'tenant_id'])
    return {'firewall_policy': body}


class ListFirewallPolicy(neutronv20.ListCommand):
    """List firewall policies that belong to a given tenant."""

    resource = 'firewall_policy'
    list_columns = ['id', 'name', 'firewall_rules']
    _formatters = {'firewall_rules': _format_firewall_rules,
                   }
    pagination_support = True
    sorting_support = True


class ShowFirewallPolicy(neutronv20.ShowCommand):
    """Show information of a given firewall policy."""

    resource = 'firewall_policy'


class CreateFirewallPolicy(neutronv20.CreateCommand):
    """Create a firewall policy."""

    resource = 'firewall_policy'

    def add_known_arguments(self, parser):
        parser.add_argument(
            'name',
            metavar='NAME',
            help=_('Name for the firewall policy.'))
        parser.add_argument(
            '--shared',
            action='store_true',
            help=_('Create a shared policy.'),
            default=argparse.SUPPRESS)
        parser.add_argument(
            '--audited',
            action='store_true',
            help=_('Sets audited to True.'),
            default=argparse.SUPPRESS)
        add_common_args(parser)

    def args2body(self, parsed_args):
        return parse_common_args(self.get_client(), parsed_args)


class UpdateFirewallPolicy(neutronv20.UpdateCommand):
    """Update a given firewall policy."""

    resource = 'firewall_policy'

    def add_known_arguments(self, parser):
        add_common_args(parser)
        parser.add_argument(
            '--name',
            help=_('Name for the firewall policy.'))
        utils.add_boolean_argument(
            parser, '--shared',
            help=_('Update the sharing status of the policy. '
                   '(True means shared).'))
        utils.add_boolean_argument(
            parser, '--audited',
            help=_('Update the audit status of the policy. '
                   '(True means auditing is enabled).'))

    def args2body(self, parsed_args):
        return parse_common_args(self.get_client(), parsed_args)


class DeleteFirewallPolicy(neutronv20.DeleteCommand):
    """Delete a given firewall policy."""

    resource = 'firewall_policy'


class FirewallPolicyInsertRule(neutronv20.UpdateCommand):
    """Insert a rule into a given firewall policy."""

    resource = 'firewall_policy'

    def call_api(self, neutron_client, firewall_policy_id, body):
        return neutron_client.firewall_policy_insert_rule(firewall_policy_id,
                                                          body)

    def args2body(self, parsed_args):
        _rule = ''
        if parsed_args.firewall_rule_id:
            _rule = neutronv20.find_resourceid_by_name_or_id(
                self.get_client(), 'firewall_rule',
                parsed_args.firewall_rule_id)
        _insert_before = ''
        if 'insert_before' in parsed_args:
            if parsed_args.insert_before:
                _insert_before = neutronv20.find_resourceid_by_name_or_id(
                    self.get_client(), 'firewall_rule',
                    parsed_args.insert_before)
        _insert_after = ''
        if 'insert_after' in parsed_args:
            if parsed_args.insert_after:
                _insert_after = neutronv20.find_resourceid_by_name_or_id(
                    self.get_client(), 'firewall_rule',
                    parsed_args.insert_after)
        body = {'firewall_rule_id': _rule,
                'insert_before': _insert_before,
                'insert_after': _insert_after}
        return body

    def get_parser(self, prog_name):
        parser = super(FirewallPolicyInsertRule, self).get_parser(prog_name)
        parser.add_argument(
            '--insert-before',
            metavar='FIREWALL_RULE',
            help=_('Insert before this rule.'))
        parser.add_argument(
            '--insert-after',
            metavar='FIREWALL_RULE',
            help=_('Insert after this rule.'))
        parser.add_argument(
            'firewall_rule_id',
            metavar='FIREWALL_RULE',
            help=_('New rule to insert.'))
        self.add_known_arguments(parser)
        return parser

    def take_action(self, parsed_args):
        neutron_client = self.get_client()
        body = self.args2body(parsed_args)
        _id = neutronv20.find_resourceid_by_name_or_id(neutron_client,
                                                       self.resource,
                                                       parsed_args.id)
        self.call_api(neutron_client, _id, body)
        print((_('Inserted firewall rule in firewall policy %(id)s') %
               {'id': parsed_args.id}), file=self.app.stdout)


class FirewallPolicyRemoveRule(neutronv20.UpdateCommand):
    """Remove a rule from a given firewall policy."""

    resource = 'firewall_policy'

    def call_api(self, neutron_client, firewall_policy_id, body):
        return neutron_client.firewall_policy_remove_rule(firewall_policy_id,
                                                          body)

    def args2body(self, parsed_args):
        _rule = ''
        if parsed_args.firewall_rule_id:
            _rule = neutronv20.find_resourceid_by_name_or_id(
                self.get_client(), 'firewall_rule',
                parsed_args.firewall_rule_id)
        body = {'firewall_rule_id': _rule}
        return body

    def get_parser(self, prog_name):
        parser = super(FirewallPolicyRemoveRule, self).get_parser(prog_name)
        parser.add_argument(
            'firewall_rule_id',
            metavar='FIREWALL_RULE',
            help=_('ID or name of the firewall rule to be removed '
                   'from the policy.'))
        self.add_known_arguments(parser)
        return parser

    def take_action(self, parsed_args):
        neutron_client = self.get_client()
        body = self.args2body(parsed_args)
        _id = neutronv20.find_resourceid_by_name_or_id(neutron_client,
                                                       self.resource,
                                                       parsed_args.id)
        self.call_api(neutron_client, _id, body)
        print((_('Removed firewall rule from firewall policy %(id)s') %
               {'id': parsed_args.id}), file=self.app.stdout)



# Copyright 2016 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutronclient._i18n import _
from neutronclient.common import exceptions
from neutronclient.common import utils
from neutronclient.common import validators
from neutronclient.neutron import v2_0 as neutronv20


def get_bgp_peer_id(client, id_or_name):
    return neutronv20.find_resourceid_by_name_or_id(client,
                                                    'bgp_peer',
                                                    id_or_name)


def validate_peer_attributes(parsed_args):
    # Validate AS number
    validators.validate_int_range(parsed_args, 'remote_as',
                                  neutronv20.bgp.speaker.MIN_AS_NUM,
                                  neutronv20.bgp.speaker.MAX_AS_NUM)
    # Validate password
    if parsed_args.auth_type != 'none' and parsed_args.password is None:
        raise exceptions.CommandError(_('Must provide password if auth-type '
                                        'is specified.'))
    if parsed_args.auth_type == 'none' and parsed_args.password:
        raise exceptions.CommandError(_('Must provide auth-type if password '
                                        'is specified.'))


class ListPeers(neutronv20.ListCommand):
    """List BGP peers."""

    resource = 'bgp_peer'
    list_columns = ['id', 'name', 'peer_ip', 'remote_as']
    pagination_support = True
    sorting_support = True


class ShowPeer(neutronv20.ShowCommand):
    """Show information of a given BGP peer."""

    resource = 'bgp_peer'


class CreatePeer(neutronv20.CreateCommand):
    """Create a BGP Peer."""

    resource = 'bgp_peer'

    def add_known_arguments(self, parser):
        parser.add_argument(
            'name',
            metavar='NAME',
            help=_('Name of the BGP peer to create.'))
        parser.add_argument(
            '--peer-ip',
            metavar='PEER_IP_ADDRESS',
            required=True,
            help=_('Peer IP address.'))
        parser.add_argument(
            '--remote-as',
            required=True,
            metavar='PEER_REMOTE_AS',
            help=_('Peer AS number. (Integer in [%(min_val)s, %(max_val)s] '
                   'is allowed.)') %
            {'min_val': neutronv20.bgp.speaker.MIN_AS_NUM,
             'max_val': neutronv20.bgp.speaker.MAX_AS_NUM})
        parser.add_argument(
            '--auth-type',
            metavar='PEER_AUTH_TYPE',
            choices=['none', 'md5'],
            default='none',
            type=utils.convert_to_lowercase,
            help=_('Authentication algorithm. Supported algorithms: '
                   'none(default), md5'))
        parser.add_argument(
            '--password',
            metavar='AUTH_PASSWORD',
            help=_('Authentication password.'))

    def args2body(self, parsed_args):
        body = {}
        validate_peer_attributes(parsed_args)
        neutronv20.update_dict(parsed_args, body,
                               ['name', 'peer_ip',
                                'remote_as', 'auth_type', 'password'])
        return {self.resource: body}


class UpdatePeer(neutronv20.UpdateCommand):
    """Update BGP Peer's information."""

    resource = 'bgp_peer'

    def add_known_arguments(self, parser):
        parser.add_argument(
            '--name',
            help=_('Updated name of the BGP peer.'))
        parser.add_argument(
            '--password',
            metavar='AUTH_PASSWORD',
            help=_('Updated authentication password.'))

    def args2body(self, parsed_args):
        body = {}
        neutronv20.update_dict(parsed_args, body, ['name', 'password'])
        return {self.resource: body}


class DeletePeer(neutronv20.DeleteCommand):
    """Delete a BGP peer."""

    resource = 'bgp_peer'



# Copyright (c) 2017 Huawei Technologies India Pvt.Limited.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from osc_lib.cli import parseractions
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils
from osc_lib.utils import columns as column_util

from neutronclient._i18n import _

LOG = logging.getLogger(__name__)

resource = 'port_chain'

_attr_map = (
    ('id', 'ID', column_util.LIST_BOTH),
    ('name', 'Name', column_util.LIST_BOTH),
    ('port_pair_groups', 'Port Pair Groups', column_util.LIST_BOTH),
    ('flow_classifiers', 'Flow Classifiers',
     column_util.LIST_BOTH),
    ('chain_parameters', 'Chain Parameters',
     column_util.LIST_BOTH),
    ('description', 'Description', column_util.LIST_LONG_ONLY),
    ('project_id', 'Project', column_util.LIST_LONG_ONLY),
)

_attr_map_dict = {
    'id': 'ID',
    'name': 'Name',
    'port_pair_groups': 'Port Pair Groups',
    'flow_classifiers': 'Flow Classifiers',
    'chain_parameters': 'Chain Parameters',
    'description': 'Description',
    'tenant_id': 'Project',
    'project_id': 'Project',
}


class CreateSfcPortChain(command.ShowOne):
    _description = _("Create a port chain")

    def get_parser(self, prog_name):
        parser = super(CreateSfcPortChain, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('Name of the port chain'))
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description for the port chain'))
        parser.add_argument(
            '--flow-classifier',
            default=[],
            metavar='<flow-classifier>',
            dest='flow_classifiers',
            action='append',
            help=_('Add flow classifier (name or ID). '
                   'This option can be repeated.'))
        parser.add_argument(
            '--chain-parameters',
            metavar='correlation=<correlation-type>,symmetric=<boolean>',
            action=parseractions.MultiKeyValueAction,
            optional_keys=['correlation', 'symmetric'],
            help=_('Dictionary of chain parameters. Supports '
                   'correlation=(mpls|nsh) (default is mpls) '
                   'and symmetric=(true|false).'))
        parser.add_argument(
            '--port-pair-group',
            metavar='<port-pair-group>',
            dest='port_pair_groups',
            required=True,
            action='append',
            help=_('Add port pair group (name or ID). '
                   'This option can be repeated.'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        attrs = _get_common_attrs(self.app.client_manager, parsed_args)
        obj = client.create_sfc_port_chain(**attrs)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, _attr_map_dict, ['location', 'tenant_id'])
        data = utils.get_dict_properties(obj, columns)
        return display_columns, data


class DeleteSfcPortChain(command.Command):
    _description = _("Delete a given port chain")

    def get_parser(self, prog_name):
        parser = super(DeleteSfcPortChain, self).get_parser(prog_name)
        parser.add_argument(
            'port_chain',
            metavar="<port-chain>",
            nargs='+',
            help=_("Port chain(s) to delete (name or ID)")
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        result = 0
        for pc in parsed_args.port_chain:
            try:
                pc_id = client.find_sfc_port_chain(
                    pc, ignore_missing=False)['id']
                client.delete_sfc_port_chain(pc_id)
            except Exception as e:
                result += 1
                LOG.error(_("Failed to delete port chain with name "
                            "or ID '%(pc)s': %(e)s"), {'pc': pc, 'e': e})
        if result > 0:
            total = len(parsed_args.port_chain)
            msg = (_("%(result)s of %(total)s port chain(s) "
                     "failed to delete.") % {'result': result,
                                             'total': total})
            raise exceptions.CommandError(msg)


class ListSfcPortChain(command.Lister):
    _description = _("List port chains")

    def get_parser(self, prog_name):
        parser = super(ListSfcPortChain, self).get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_("List additional fields in output")
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        data = client.sfc_port_chains()
        headers, columns = column_util.get_column_definitions(
            _attr_map, long_listing=parsed_args.long)
        return (headers,
                (utils.get_dict_properties(s, columns)
                 for s in data))


class SetSfcPortChain(command.Command):
    _description = _("Set port chain properties")

    def get_parser(self, prog_name):
        parser = super(SetSfcPortChain, self).get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Name of the port chain'))
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description for the port chain'))
        parser.add_argument(
            '--flow-classifier',
            metavar='<flow-classifier>',
            dest='flow_classifiers',
            action='append',
            help=_('Add flow classifier (name or ID). '
                   'This option can be repeated.'))
        parser.add_argument(
            '--no-flow-classifier',
            action='store_true',
            help=_('Remove associated flow classifiers from the port chain'))
        parser.add_argument(
            '--port-pair-group',
            metavar='<port-pair-group>',
            dest='port_pair_groups',
            action='append',
            help=_('Add port pair group (name or ID). '
                   'Current port pair groups order is kept, the added port '
                   'pair group will be placed at the end of the port chain. '
                   'This option can be repeated.'))
        parser.add_argument(
            '--no-port-pair-group',
            action='store_true',
            help=_('Remove associated port pair groups from the port chain. '
                   'At least one --port-pair-group must be specified '
                   'together.'))
        parser.add_argument(
            'port_chain',
            metavar='<port-chain>',
            help=_("Port chain to modify (name or ID)"))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        pc_id = client.find_sfc_port_chain(parsed_args.port_chain,
                                           ignore_missing=False)['id']
        attrs = _get_common_attrs(self.app.client_manager, parsed_args,
                                  is_create=False)
        if parsed_args.no_flow_classifier:
            attrs['flow_classifiers'] = []
        if parsed_args.flow_classifiers:
            if parsed_args.no_flow_classifier:
                fc_list = []
            else:
                fc_list = client.find_sfc_port_chain(
                    parsed_args.port_chain,
                    ignore_missing=False
                )['flow_classifiers']
            for fc in parsed_args.flow_classifiers:
                fc_id = client.find_sfc_flow_classifier(
                    fc,
                    ignore_missing=False)['id']
                if fc_id not in fc_list:
                    fc_list.append(fc_id)
            attrs['flow_classifiers'] = fc_list
        if (parsed_args.no_port_pair_group and not
                parsed_args.port_pair_groups):
            message = _('At least one --port-pair-group must be specified.')
            raise exceptions.CommandError(message)
        if parsed_args.no_port_pair_group and parsed_args.port_pair_groups:
            ppg_list = []
            for ppg in parsed_args.port_pair_groups:
                ppg_id = client.find_sfc_port_pair_group(
                    ppg, ignore_missing=False)['id']
                if ppg_id not in ppg_list:
                    ppg_list.append(ppg_id)
            attrs['port_pair_groups'] = ppg_list
        if (parsed_args.port_pair_groups and
                not parsed_args.no_port_pair_group):
            ppg_list = client.find_sfc_port_chain(
                parsed_args.port_chain,
                ignore_missing=False
            )['port_pair_groups']
            for ppg in parsed_args.port_pair_groups:
                ppg_id = client.find_sfc_port_pair_group(
                    ppg, ignore_missing=False)['id']
                if ppg_id not in ppg_list:
                    ppg_list.append(ppg_id)
            attrs['port_pair_groups'] = ppg_list
        try:
            client.update_sfc_port_chain(pc_id, **attrs)
        except Exception as e:
            msg = (_("Failed to update port chain '%(pc)s': %(e)s")
                   % {'pc': parsed_args.port_chain, 'e': e})
            raise exceptions.CommandError(msg)


class ShowSfcPortChain(command.ShowOne):
    _description = _("Display port chain details")

    def get_parser(self, prog_name):
        parser = super(ShowSfcPortChain, self).get_parser(prog_name)
        parser.add_argument(
            'port_chain',
            metavar="<port-chain>",
            help=_("Port chain to display (name or ID)")
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        pc_id = client.find_sfc_port_chain(parsed_args.port_chain,
                                           ignore_missing=False)['id']
        obj = client.get_sfc_port_chain(pc_id)
        display_columns, columns = utils.get_osc_show_columns_for_sdk_resource(
            obj, _attr_map_dict, ['location', 'tenant_id'])
        data = utils.get_dict_properties(obj, columns)
        return display_columns, data


class UnsetSfcPortChain(command.Command):
    _description = _("Unset port chain properties")

    def get_parser(self, prog_name):
        parser = super(UnsetSfcPortChain, self).get_parser(prog_name)
        parser.add_argument(
            'port_chain',
            metavar='<port-chain>',
            help=_("Port chain to unset (name or ID)"))
        port_chain = parser.add_mutually_exclusive_group()
        port_chain.add_argument(
            '--flow-classifier',
            action='append',
            metavar='<flow-classifier>',
            dest='flow_classifiers',
            help=_('Remove flow classifier(s) from the port chain '
                   '(name or ID). This option can be repeated.'))
        port_chain.add_argument(
            '--all-flow-classifier',
            action='store_true',
            help=_('Remove all flow classifiers from the port chain'))
        parser.add_argument(
            '--port-pair-group',
            metavar='<port-pair-group>',
            dest='port_pair_groups',
            action='append',
            help=_('Remove port pair group(s) from the port chain '
                   '(name or ID). This option can be repeated.'))
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        pc_id = client.find_sfc_port_chain(parsed_args.port_chain,
                                           ignore_missing=False)['id']
        attrs = {}
        if parsed_args.flow_classifiers:
            fc_list = client.find_sfc_port_chain(
                parsed_args.port_chain, ignore_missing=False
            )['flow_classifiers']
            for fc in parsed_args.flow_classifiers:
                fc_id = client.find_sfc_flow_classifier(
                    fc,
                    ignore_missing=False)['id']
                if fc_id in fc_list:
                    fc_list.remove(fc_id)
            attrs['flow_classifiers'] = fc_list
        if parsed_args.all_flow_classifier:
            attrs['flow_classifiers'] = []
        if parsed_args.port_pair_groups:
            ppg_list = client.find_sfc_port_chain(
                parsed_args.port_chain,
                ignore_missing=False)['port_pair_groups']
            for ppg in parsed_args.port_pair_groups:
                ppg_id = client.find_sfc_port_pair_group(
                    ppg,
                    ignore_missing=False)['id']
                if ppg_id in ppg_list:
                    ppg_list.remove(ppg_id)
            if ppg_list == []:
                message = _('At least one port pair group must be'
                            ' specified.')
                raise exceptions.CommandError(message)
            attrs['port_pair_groups'] = ppg_list
        try:
            client.update_sfc_port_chain(pc_id, **attrs)
        except Exception as e:
            msg = (_("Failed to unset port chain '%(pc)s': %(e)s")
                   % {'pc': parsed_args.port_chain, 'e': e})
            raise exceptions.CommandError(msg)


def _get_common_attrs(client_manager, parsed_args, is_create=True):
    attrs = {}
    client = client_manager.network
    if parsed_args.name is not None:
        attrs['name'] = parsed_args.name
    if parsed_args.description is not None:
        attrs['description'] = parsed_args.description
    if parsed_args.port_pair_groups:
        attrs['port_pair_groups'] = [client.find_sfc_port_pair_group(
            ppg, ignore_missing=False)['id']
                                     for ppg in parsed_args.port_pair_groups]
    if parsed_args.flow_classifiers:
        attrs['flow_classifiers'] = [client.find_sfc_flow_classifier(
            fc, ignore_missing=False)['id']
                                     for fc in parsed_args.flow_classifiers]
    if is_create is True:
        _get_attrs(attrs, parsed_args)
    return attrs


def _get_attrs(attrs, parsed_args):
    if parsed_args.chain_parameters is not None:
        chain_params = {}
        for chain_param in parsed_args.chain_parameters:
            if 'correlation' in chain_param:
                chain_params['correlation'] = chain_param['correlation']
            if 'symmetric' in chain_param:
                chain_params['symmetric'] = chain_param['symmetric']
        attrs['chain_parameters'] = chain_params



# Copyright 2013 Mirantis Inc.
# Copyright 2014 Blue Box Group, Inc.
# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Blue Box, an IBM Company
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutronclient._i18n import _
from neutronclient.common import exceptions
from neutronclient.common import utils
from neutronclient.neutron import v2_0 as neutronV20


def _get_loadbalancer_id(client, lb_id_or_name):
    return neutronV20.find_resourceid_by_name_or_id(
        client, 'loadbalancer', lb_id_or_name,
        cmd_resource='lbaas_loadbalancer')


def _get_listener(client, listener_id_or_name):
    return neutronV20.find_resource_by_name_or_id(
        client, 'listener', listener_id_or_name)


def _get_listener_id(client, listener_id_or_name):
    return neutronV20.find_resourceid_by_name_or_id(
        client, 'listener', listener_id_or_name)


def _add_common_args(parser, is_create=True):
    parser.add_argument(
        '--description',
        help=_('Description of the pool.'))
    parser.add_argument(
        '--name', help=_('The name of the pool.'))
    parser.add_argument(
        '--lb-algorithm',
        required=is_create,
        type=utils.convert_to_uppercase,
        choices=['ROUND_ROBIN', 'LEAST_CONNECTIONS', 'SOURCE_IP'],
        help=_('The algorithm used to distribute load between the members '
               'of the pool.'))


def _parse_common_args(parsed_args):
    body = {}
    neutronV20.update_dict(parsed_args,
                           body, ['description', 'lb_algorithm', 'name'])
    return body


class ListPool(neutronV20.ListCommand):
    """LBaaS v2 List pools that belong to a given tenant."""

    resource = 'pool'
    shadow_resource = 'lbaas_pool'
    list_columns = ['id', 'name', 'lb_algorithm', 'protocol',
                    'admin_state_up']
    pagination_support = True
    sorting_support = True


class ShowPool(neutronV20.ShowCommand):
    """LBaaS v2 Show information of a given pool."""

    resource = 'pool'
    shadow_resource = 'lbaas_pool'

    def cleanup_output_data(self, data):
        if 'members' not in data['pool']:
            return []
        member_info = []
        for member in data['pool']['members']:
            member_info.append(member['id'])
        data['pool']['members'] = member_info


class CreatePool(neutronV20.CreateCommand):
    """LBaaS v2 Create a pool."""

    resource = 'pool'
    shadow_resource = 'lbaas_pool'

    def add_known_arguments(self, parser):
        _add_common_args(parser)
        parser.add_argument(
            '--admin-state-down',
            dest='admin_state', action='store_false',
            help=_('Set admin state up to false.'))
        parser.add_argument(
            '--listener',
            help=_('Listener whose default-pool should be set to this pool. '
                   'At least one of --listener or --loadbalancer must be '
                   'specified.'))
        parser.add_argument(
            '--loadbalancer',
            help=_('Loadbalancer with which this pool should be associated. '
                   'At least one of --listener or --loadbalancer must be '
                   'specified.'))
        parser.add_argument(
            '--protocol',
            type=utils.convert_to_uppercase,
            required=True,
            choices=['HTTP', 'HTTPS', 'TCP'],
            help=_('Protocol for balancing.'))
        parser.add_argument(
            '--session-persistence',
            metavar='type=TYPE[,cookie_name=COOKIE_NAME]',
            type=utils.str2dict_type(required_keys=['type'],
                                     optional_keys=['cookie_name']),
            help=_('The type of session persistence to use and associated '
                   'cookie name.'))

    def args2body(self, parsed_args):
        if not parsed_args.listener and not parsed_args.loadbalancer:
            message = _('At least one of --listener or --loadbalancer must be '
                        'specified.')
            raise exceptions.CommandError(message)
        body = _parse_common_args(parsed_args)
        if parsed_args.listener:
            listener_id = _get_listener_id(
                self.get_client(),
                parsed_args.listener)
            body['listener_id'] = listener_id
        if parsed_args.loadbalancer:
            loadbalancer_id = _get_loadbalancer_id(
                self.get_client(),
                parsed_args.loadbalancer)
            body['loadbalancer_id'] = loadbalancer_id
        body['admin_state_up'] = parsed_args.admin_state
        neutronV20.update_dict(parsed_args, body,
                               ['tenant_id', 'protocol',
                                'session_persistence'])
        return {self.resource: body}


class UpdatePool(neutronV20.UpdateCommand):
    """LBaaS v2 Update a given pool."""

    resource = 'pool'
    shadow_resource = 'lbaas_pool'

    def add_known_arguments(self, parser):
        utils.add_boolean_argument(
            parser, '--admin-state-up',
            help=_('Update the administrative state of '
                   'the pool (True meaning "Up").'))
        session_group = parser.add_mutually_exclusive_group()
        session_group.add_argument(
            '--session-persistence',
            metavar='type=TYPE[,cookie_name=COOKIE_NAME]',
            type=utils.str2dict_type(required_keys=['type'],
                                     optional_keys=['cookie_name']),
            help=_('The type of session persistence to use and associated '
                   'cookie name.'))
        session_group.add_argument(
            '--no-session-persistence',
            action='store_true',
            help=_('Clear session persistence for the pool.'))
        _add_common_args(parser, False)

    def args2body(self, parsed_args):
        body = _parse_common_args(parsed_args)
        if parsed_args.no_session_persistence:
            body['session_persistence'] = None
        elif parsed_args.session_persistence:
            body['session_persistence'] = parsed_args.session_persistence
        neutronV20.update_dict(parsed_args, body,
                               ['admin_state_up'])
        return {self.resource: body}


class DeletePool(neutronV20.DeleteCommand):
    """LBaaS v2 Delete a given pool."""

    resource = 'pool'
    shadow_resource = 'lbaas_pool'
